"""
This file will contain all of the specific post-processing functionalities for water-energy-food-environment (WEFE)
systems e.g. calculating the crop yield and the total water footprint
"""


def calculate_crop_yield():
    return


def calculate_total_water_footprint():
    return
