"""
This subpackage of wefe will contain the global global_specs that can be used across multiple projects
e.g. different PV panel global_specs or different crop global_specs (energy content etc)
ToDo: decide if these global_specs actually belong in oemof-tabular-plugins or if they should be stored somewhere else
"""

from .crop_specs import crop_dict
from .soil_specs import soil_dict
